==> llm_utility_pack-1.2-py3-none-any.whl: utility_pack/vector_storage_helper.py <==
from concurrent.futures import ThreadPoolExecutor, as_completed
import lmdb, struct, pickle, os, atexit
import numpy as np

class LmdbStorage:
    def __init__(self, path, map_size=20*1024*1024*1024): # e.g. 20GB by default
        self.env = lmdb.open(path, map_size=map_size)
        atexit.register(self.close)

    def _int_to_bytes(self, x):
        return struct.pack('<q', x)

    def store_data(self, data, identifiers, batch_size=5000):
        total = len(data)
        with self.env.begin(write=True) as txn:
            for i in range(0, total, batch_size):
                batch_data = data[i:i+batch_size]
                batch_ids = identifiers[i:i+batch_size]
                with txn.cursor() as curs:
                    for vec, id in zip(batch_data, batch_ids):
                        curs.put(
                            self._int_to_bytes(id) if isinstance(id, int) else id.encode(),
                            pickle.dumps(vec)
                        )

    def get_data(self, identifiers):
        datas = []
        with self.env.begin() as txn:
            for id in identifiers:
                data = txn.get(self._int_to_bytes(id) if isinstance(id, int) else id.encode())
                if data:
                    datas.append(pickle.loads(data))
                else:
                    datas.append(None)
        return [ v for v in datas if v is not None ]
        
    def store_vectors(self, data, identifiers, batch_size=5000):
        total = len(data)
        with self.env.begin(write=True) as txn:
            for i in range(0, total, batch_size):
                batch_data = data[i:i+batch_size]
                batch_ids = identifiers[i:i+batch_size]
                with txn.cursor() as curs:
                    for vec, id in zip(batch_data, batch_ids):
                        curs.put(
                            self._int_to_bytes(id) if isinstance(id, int) else id.encode(),
                            np.array(vec, dtype=np.float32).tobytes()
                        )
    
    def get_vectors(self, identifiers):
        datas = []
        with self.env.begin() as txn:
            for id in identifiers:
                data = txn.get(self._int_to_bytes(id) if isinstance(id, int) else id.encode())
                if data:
                    datas.append(np.frombuffer(data, dtype=np.float32))
                else:
                    datas.append(None)
        return [v for v in datas if v is not None]

    def delete_data(self, identifiers):
        with self.env.begin(write=True) as txn:
            for id in identifiers:
                txn.delete(self._int_to_bytes(id) if isinstance(id, int) else id.encode())

    def get_data_count(self):
        with self.env.begin() as txn:
            return txn.stat()['entries']

    def sync(self):
        self.env.sync()

    def close(self):
        self.env.close()

class ShardedLmdbStorage:
    """
    A sharded wrapper for LmdbStorage that splits data across multiple shards.
    Each shard is an instance of LmdbStorage stored in a subdirectory under a base path.
    """
    def __init__(self, base_path, num_shards, map_size=70*1024*1024*1024):
        """
        :param base_path: Base directory where shard subdirectories will be created.
        :param num_shards: Number of shards.
        :param map_size: Map size for each LMDB environment.
        """
        self.num_shards = num_shards
        self.shards = {}
        for shard_idx in range(num_shards):
            shard_path = os.path.join(base_path, f"shard_{shard_idx}")
            os.makedirs(shard_path, exist_ok=True)
            self.shards[shard_idx] = LmdbStorage(shard_path, map_size=map_size)

    def _get_shard_for_id(self, identifier):
        """
        Determine the shard index for a given identifier.
        Uses modulo arithmetic for integers, and hash for non-integers.
        """
        if isinstance(identifier, int):
            return identifier % self.num_shards
        else:
            # Using Python's built-in hash. For more consistency across runs,
            # you might want to use a stable hash function like hashlib.
            return hash(identifier) % self.num_shards

    def store_data(self, data, identifiers, batch_size=5000):
        """
        Stores data items by grouping them by shard.
        """
        # Group data and identifiers by shard index
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        # Call store_data on each shard that has items
        for shard, lmdb_storage in self.shards.items():
            if shard_data[shard]:
                lmdb_storage.store_data(shard_data[shard], shard_ids[shard], batch_size=batch_size)
    
    def get_data(self, identifiers):
        """
        Retrieves data items by grouping identifiers by shard.
        Returns a combined list of found data.
        """
        results = []
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_ids[shard]:
                results.extend(lmdb_storage.get_data(shard_ids[shard]))
        return results

    def store_vectors(self, data, identifiers, batch_size=5000):
        """
        Stores vector data by grouping them by shard.
        """
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_data[shard]:
                lmdb_storage.store_vectors(shard_data[shard], shard_ids[shard], batch_size=batch_size)

    def get_vectors(self, identifiers):
        """
        Retrieves vector data items by grouping identifiers by shard.
        """
        results = []
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_ids[shard]:
                results.extend(lmdb_storage.get_vectors(shard_ids[shard]))
        return results

    def delete_data(self, identifiers):
        """
        Deletes data items by grouping identifiers by shard.
        """
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_ids[shard]:
                lmdb_storage.delete_data(shard_ids[shard])

    def get_data_count(self):
        """
        Returns the total count of entries across all shards.
        """
        total = 0
        for lmdb_storage in self.shards.values():
            total += lmdb_storage.get_data_count()
        return total

    def sync(self):
        """
        Synchronizes all LMDB environments.
        """
        for lmdb_storage in self.shards.values():
            lmdb_storage.sync()

    def close(self):
        """
        Closes all LMDB environments.
        """
        for lmdb_storage in self.shards.values():
            lmdb_storage.close()

    # Parallelized versions of the methods
    def store_data_parallel(self, data, identifiers, batch_size=5000, max_workers=None):
        """
        Stores data items by grouping them by shard, using parallel execution.
        """
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for shard, lmdb_storage in self.shards.items():
                if shard_data[shard]:
                    futures.append(executor.submit(lmdb_storage.store_data, shard_data[shard], shard_ids[shard], batch_size))
            for future in as_completed(futures):
                future.result()  # Ensure all futures complete

    def get_data_parallel(self, identifiers, max_workers=None):
        """
        Retrieves data items by grouping identifiers by shard, using parallel execution.
        """
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        
        results = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(self.shards[shard].get_data, shard_ids[shard]): shard for shard in shard_ids if shard_ids[shard]}
            for future in as_completed(futures):
                results.extend(future.result())
        return results

    def store_vectors_parallel(self, data, identifiers, batch_size=5000, max_workers=None):
        """
        Stores vector data by grouping them by shard, using parallel execution.
        """
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for shard, lmdb_storage in self.shards.items():
                if shard_data[shard]:
                    futures.append(executor.submit(lmdb_storage.store_vectors, shard_data[shard], shard_ids[shard], batch_size))
            for future in as_completed(futures):
                future.result()  # Ensure all futures complete

    def get_vectors_parallel(self, identifiers, max_workers=None):
        """
        Retrieves vector data items by grouping identifiers by shard, using parallel execution.
        """
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        
        results = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(self.shards[shard].get_vectors, shard_ids[shard]): shard for shard in shard_ids if shard_ids[shard]}
            for future in as_completed(futures):
                results.extend(future.result())
        return results

==> llm_utility_pack-2.1-py3-none-any.whl: utility_pack/vector_storage_helper.py <==
from concurrent.futures import ThreadPoolExecutor, as_completed
import lmdb, struct, pickle, os, atexit, hashlib
import numpy as np

class LmdbStorage:
    def __init__(self, path, map_size=20*1024*1024*1024): # e.g. 20GB by default
        self.env = lmdb.open(path, map_size=map_size)
        atexit.register(self.close)

    def _int_to_bytes(self, x):
        return struct.pack('<q', x)

    def store_data(self, data, identifiers, batch_size=5000):
        total = len(data)
        with self.env.begin(write=True) as txn:
            for i in range(0, total, batch_size):
                batch_data = data[i:i+batch_size]
                batch_ids = identifiers[i:i+batch_size]
                with txn.cursor() as curs:
                    for vec, id in zip(batch_data, batch_ids):
                        curs.put(
                            self._int_to_bytes(id) if isinstance(id, int) else id.encode(),
                            pickle.dumps(vec)
                        )

    def get_data(self, identifiers):
        datas = []
        with self.env.begin() as txn:
            for id in identifiers:
                data = txn.get(self._int_to_bytes(id) if isinstance(id, int) else id.encode())
                if data:
                    datas.append(pickle.loads(data))
                else:
                    datas.append(None)
        return [ v for v in datas if v is not None ]
        
    def store_vectors(self, data, identifiers, batch_size=5000):
        total = len(data)
        with self.env.begin(write=True) as txn:
            for i in range(0, total, batch_size):
                batch_data = data[i:i+batch_size]
                batch_ids = identifiers[i:i+batch_size]
                with txn.cursor() as curs:
                    for vec, id in zip(batch_data, batch_ids):
                        curs.put(
                            self._int_to_bytes(id) if isinstance(id, int) else id.encode(),
                            np.array(vec, dtype=np.float32).tobytes()
                        )
    
    def get_vectors(self, identifiers):
        datas = []
        with self.env.begin() as txn:
            for id in identifiers:
                data = txn.get(self._int_to_bytes(id) if isinstance(id, int) else id.encode())
                if data:
                    datas.append(np.frombuffer(data, dtype=np.float32))
                else:
                    datas.append(None)
        return [v for v in datas if v is not None]

    def delete_data(self, identifiers):
        with self.env.begin(write=True) as txn:
            for id in identifiers:
                txn.delete(self._int_to_bytes(id) if isinstance(id, int) else id.encode())

    def get_data_count(self):
        with self.env.begin() as txn:
            return txn.stat()['entries']

    def sync(self):
        self.env.sync()

    def close(self):
        self.env.close()

class ShardedLmdbStorage:
    """
    A sharded wrapper for LmdbStorage that splits data across multiple shards.
    Each shard is an instance of LmdbStorage stored in a subdirectory under a base path.
    """
    def __init__(self, base_path, num_shards=5, map_size=70*1024*1024*1024):
        """
        :param base_path: Base directory where shard subdirectories will be created.
        :param num_shards: Number of shards.
        :param map_size: Map size for each LMDB environment.
        """
        self.num_shards = num_shards
        self.shards = {}
        for shard_idx in range(num_shards):
            shard_path = os.path.join(base_path, f"shard_{shard_idx}")
            os.makedirs(shard_path, exist_ok=True)
            self.shards[shard_idx] = LmdbStorage(shard_path, map_size=map_size)

    def _get_shard_for_id(self, identifier):
        identifier_str = str(identifier).encode('utf-8')
        h = hashlib.md5(identifier_str).hexdigest()
        return int(h, 16) % self.num_shards

    def store_data(self, data, identifiers, batch_size=5000):
        """
        Stores data items by grouping them by shard.
        """
        # Group data and identifiers by shard index
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        # Call store_data on each shard that has items
        for shard, lmdb_storage in self.shards.items():
            if shard_data[shard]:
                lmdb_storage.store_data(shard_data[shard], shard_ids[shard], batch_size=batch_size)
    
    def get_data(self, identifiers):
        """
        Retrieves data items by grouping identifiers by shard.
        Returns a combined list of found data.
        """
        results = []
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_ids[shard]:
                results.extend(lmdb_storage.get_data(shard_ids[shard]))
        return results

    def store_vectors(self, data, identifiers, batch_size=5000):
        """
        Stores vector data by grouping them by shard.
        """
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_data[shard]:
                lmdb_storage.store_vectors(shard_data[shard], shard_ids[shard], batch_size=batch_size)

    def get_vectors(self, identifiers):
        """
        Retrieves vector data items by grouping identifiers by shard.
        """
        results = []
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_ids[shard]:
                results.extend(lmdb_storage.get_vectors(shard_ids[shard]))
        return results

    def delete_data(self, identifiers):
        """
        Deletes data items by grouping identifiers by shard.
        """
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        for shard, lmdb_storage in self.shards.items():
            if shard_ids[shard]:
                lmdb_storage.delete_data(shard_ids[shard])

    def get_data_count(self):
        """
        Returns the total count of entries across all shards.
        """
        total = 0
        for lmdb_storage in self.shards.values():
            total += lmdb_storage.get_data_count()
        return total

    def sync(self):
        """
        Synchronizes all LMDB environments.
        """
        for lmdb_storage in self.shards.values():
            lmdb_storage.sync()

    def close(self):
        """
        Closes all LMDB environments.
        """
        for lmdb_storage in self.shards.values():
            lmdb_storage.close()

    # Parallelized versions of the methods
    def store_data_parallel(self, data, identifiers, batch_size=5000, max_workers=None):
        """
        Stores data items by grouping them by shard, using parallel execution.
        """
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for shard, lmdb_storage in self.shards.items():
                if shard_data[shard]:
                    futures.append(executor.submit(lmdb_storage.store_data, shard_data[shard], shard_ids[shard], batch_size))
            for future in as_completed(futures):
                future.result()  # Ensure all futures complete

    def get_data_parallel(self, identifiers, max_workers=None):
        """
        Retrieves data items by grouping identifiers by shard, using parallel execution.
        """
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        
        results = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(self.shards[shard].get_data, shard_ids[shard]): shard for shard in shard_ids if shard_ids[shard]}
            for future in as_completed(futures):
                results.extend(future.result())
        return results

    def store_vectors_parallel(self, data, identifiers, batch_size=5000, max_workers=None):
        """
        Stores vector data by grouping them by shard, using parallel execution.
        """
        shard_data = {i: [] for i in range(self.num_shards)}
        shard_ids = {i: [] for i in range(self.num_shards)}
        for vec, identifier in zip(data, identifiers):
            shard = self._get_shard_for_id(identifier)
            shard_data[shard].append(vec)
            shard_ids[shard].append(identifier)
        
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for shard, lmdb_storage in self.shards.items():
                if shard_data[shard]:
                    futures.append(executor.submit(lmdb_storage.store_vectors, shard_data[shard], shard_ids[shard], batch_size))
            for future in as_completed(futures):
                future.result()  # Ensure all futures complete

    def get_vectors_parallel(self, identifiers, max_workers=None):
        """
        Retrieves vector data items by grouping identifiers by shard, using parallel execution.
        """
        shard_ids = {i: [] for i in range(self.num_shards)}
        for identifier in identifiers:
            shard = self._get_shard_for_id(identifier)
            shard_ids[shard].append(identifier)
        
        results = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(self.shards[shard].get_vectors, shard_ids[shard]): shard for shard in shard_ids if shard_ids[shard]}
            for future in as_completed(futures):
                results.extend(future.result())
        return results
